import networkx as nx
from mmodel.draw import draw_graph
from copy import deepcopy
from mmodel.filter import subnodes_by_inputs, subnodes_by_outputs
from mmodel.utility import (
    modelgraph_signature,
    modelgraph_returns,
    replace_subgraph,
    modify_node,
    parse_modifiers,
    content_wrap,
)
from mmodel.parser import parser_engine


class ModelGraph(nx.DiGraph):

    """Create model graphs.

    ModelGraph inherits from `networkx.DiGraph()`, which has all `DiGraph`
    methods.

    The class adds the "type" attribute to the graph attribute. The factory method
    returns a copy of the dictionary. It is equivalent to
    ``{"type": "ModelGraph"}.copy()`` when called.

    The additional graph operations are added:
    - add_grouped_edges and set_node_objects.
    - Method ``add_grouped_edges``, cannot have both edges list.
    - Method ``set_node_object`` updates nodes with the node callable "func" and output.
    - The method adds callable signature 'sig' to the node attribute.
    """

    graph_attr_dict_factory = {"type": "ModelGraph"}.copy

    def __init__(self, *args, parser=parser_engine, **kwargs):
        super().__init__(*args, **kwargs)
        self._parser = parser

    def set_node_object(
        self,
        node,
        func,
        output: str = None,
        inputs: list = None,
        modifiers: list = None,
    ):
        """Add or update the functions of an existing node.

        In the end, the edge attributes are re-determined
        Modifiers are applied directly onto the node. The parser checks the
        function type and returns (at least) three dictionary entries:
        _func, functype, doc.
        """

        node_dict = self.nodes[node]

        modifiers = modifiers or list()
        attr_dict = self._parser(node, func, output, inputs, modifiers)
        node_dict.update(attr_dict)

        self.update_graph()

    def set_node_objects_from(self, node_objects: list):
        """Update the functions of existing nodes.

        The method is the same as adding a node object.
        """

        for node_obj in node_objects:
            # unzipping works for input with or without modifiers
            self.set_node_object(*node_obj)

    def add_edge(self, u_of_edge, v_of_edge, **attr):
        """Modify add_edge to update the edge attribute in the end."""

        super().add_edge(u_of_edge, v_of_edge, **attr)
        self.update_graph()

    def add_edges_from(self, ebunch_to_add, **attr):
        """Modify add_edges_from to update the edge attributes."""

        super().add_edges_from(ebunch_to_add, **attr)
        self.update_graph()

    def add_grouped_edge(self, u, v):
        """Add linked edge

        For mmodel, a group edge (u, v) allows u or v
        to be a list of nodes. Represents several nodes
        flowing into one node or the other way around.
        """

        if isinstance(u, list) and isinstance(v, list):
            raise Exception("only one edge node can be a list")

        # use add edges from to run less update graph
        # currently a compromise
        if isinstance(u, list):
            self.add_edges_from([(_u, v) for _u in u])
        elif isinstance(v, list):
            self.add_edges_from([(u, _v) for _v in v])
        else:  # neither is a list
            self.add_edge(u, v)

    def add_grouped_edges_from(self, group_edges: list):
        """Add edges from grouped values."""

        for u, v in group_edges:
            self.add_grouped_edge(u, v)

    def update_graph(self):
        """Update edge attributes based on node objects and edges."""

        for u, v in self.edges:
            # the edge "var" is not defined if the parent node does not
            # have "output" attribute, or the child node does not have
            # the parameter

            # extract the parameter dictionary
            v_sig = getattr(self.nodes[v].get("sig", None), "parameters", {})
            if "output" in self.nodes[u] and self.nodes[u]["output"] in v_sig:
                self.edges[u, v]["var"] = self.nodes[u]["output"]

    def node_metadata(self, node: str, full=True, wrap_width=80):
        """Printout node metadata.

        The metadata includes the node information and the node function
        information. If the node is a ``mmodel.Model`` instance; outputs its metadata.
        """

        node_dict = self.nodes[node]

        metadata_list = [
            node,
            "",
            f"{node_dict['func'].__name__}{node_dict['sig']}",
            f"return: {node_dict['output']}",
        ]

        if full:  # adds functype, modifiers, and docs
            metadata_list.append(f"functype: {node_dict['functype']}")
            metadata_list.extend(parse_modifiers(node_dict["modifiers"]))
            doc = node_dict["doc"] or ""
            metadata_list.extend(["", doc])

        wrapped_list = content_wrap(metadata_list, width=wrap_width)

        return "\n".join(wrapped_list).rstrip()

    # graph properties
    @property
    def signature(self):
        """Graph signature.

        :rtype: inspect.Signature object
        """
        return modelgraph_signature(self)

    @property
    def returns(self):
        """Graph returns.

        :rtype: list
        """
        return modelgraph_returns(self)

    # graph operations
    def subgraph(self, nodes=None, inputs=None, outputs=None):
        """Extract subgraph by nodes, inputs, and output.

        If multiple parameters are specified, the result is a union
        of the selection.
        """

        nodes = nodes or []
        node_inputs = subnodes_by_inputs(self, inputs or [])
        node_outputs = subnodes_by_outputs(self, outputs or [])

        # convert nodes to list because the parent class method accepts generator
        # for nodes.
        # may consider not using the same name as the parent class to avoid collision
        subgraph_nodes = set(list(nodes) + node_inputs + node_outputs)  # unique nodes

        return super().subgraph(subgraph_nodes)

    def replace_subgraph(
        self, subgraph, name, func, output=None, inputs=None, modifiers=None
    ):
        """Replace subgraph with a node"""
        return replace_subgraph(self, subgraph, name, func, output, inputs, modifiers)

    def modify_node(
        self, node, func=None, output=None, inputs=None, modifiers=None, inplace=False
    ):
        """Modify node attributes"""
        return modify_node(self, node, func, output, inputs, modifiers, inplace)

    def draw(self, style="full", export=None, wrap_width=30):
        """Draw the graph.

        Draws the default styled graph.

        :param str style: there are three styles, plain, short and full.
            Plain shows nodes only, short shows part of the metadata, and
            long shows all the metadata.
        :param str export: filename to save the graph as. The file extension
            is needed.

        """

        return draw_graph(self, str(self), style, export, wrap_width)

    def deepcopy(self):
        """Deepcopy graph.

        The graph.copy method is a shallow copy. Deepcopy creates copy for the
        attributes dictionary.
        `graph.copy<https://networkx.org/documentation/stable/reference/classes
        /generated/networkx.Graph.copy.html>_`

        However, for subgraphs, deepcopy is incredibly inefficient because
        subgraph contains '_graph', which stores the original graph.
        An alternative method is to copy the code from the copy method,
        but use deepcopy for the items.
        """
        # return deepcopy(self)

        G = self.__class__()
        G.graph.update(deepcopy(self.graph))
        G.add_nodes_from((n, deepcopy(d)) for n, d in self._node.items())
        G.add_edges_from(
            (u, v, deepcopy(datadict))
            for u, nbrs in self._adj.items()
            for v, datadict in nbrs.items()
        )

        return G



import types
from mmodel.modifier import signature_modifier, pos_signature_modifier
import numpy as np
from mmodel.model import Model
import inspect


class NodeParser:
    """Output node attribute based on inputs.

    The class check function outputs the necessary properties
    of the function and add it to the node attribute.

    The class currently supports Python built-in functions, regular
    function, numpy.ufunc, and Model class instance.

    To add additional function types, subclass this class.
    """

    def __init__(self, parser_dict):
        self._parser_dict = parser_dict

    def __call__(self, node, func, output, inputs, modifiers):
        """Check if the function belongs to the correct type.

        Returns customized dictionary if that is true. It is possible
        that some of the function options overlap when the list goes
        long, the order of the parser is important.
        """
        for parser in self._parser_dict.values():
            attr_dict = parser(node, func, output, inputs, modifiers)
            if attr_dict:
                func = attr_dict["_func"]
                # apply the modifiers
                for mdf, kwargs in modifiers:
                    func = mdf(func, **kwargs)

                sig = inspect.signature(func)
                base_dict = {
                    "func": func,
                    "sig": sig,
                    "output": output,
                    "modifiers": modifiers,
                }
                # attr dict can overwrite the base dict
                attr_dict.update(base_dict)
                return attr_dict


def parse_docstring(docstring):
    """Parse docstring from built-in function and numpy.ufunc.

    The built-in and ufunv type docstring location is not consistent
    some module/function has the repr at the first line, and some don't.
    Here we try to grab the first line that starts with an upper case
    and ends with a period.
    """

    doc = ""
    for line in docstring.splitlines():
        line = line.strip()
        if line and line[0].isupper() and line.endswith("."):
            doc = line
            break
    return doc


def default_parser(node, func, output, inputs, modifiers):
    """Return the default function dictionary.

    The function needs to have the start uppercase and end period style of
    docstring.
    """
    if callable(func):
        func_dict = {}

        doc = ""
        if hasattr(func, "__doc__") and func.__doc__:
            doc = parse_docstring(func.__doc__)

        func_dict['doc'] = doc

        if inputs:
            func = signature_modifier(func, inputs)

        func_dict.update({"_func": func, "functype": "callable"})
        return func_dict

    else:
        raise Exception(f"Node {repr(node)} has invalid function type.")


def builtin_parser(node, func, output, inputs, modifiers):
    """Check if the function is a built-in function."""
    if isinstance(func, types.BuiltinFunctionType):

        doc = parse_docstring(func.__doc__)

        if inputs:
            func = pos_signature_modifier(func, inputs)
        else:
            raise Exception(
                f"Node {repr(node)} built-in type function "
                "requires 'inputs' definition."
            )

        return {"_func": func, "functype": "builtin", "doc": doc}


def ufunc_parser(node, func, output, inputs, modifiers):
    """Check if the function is a numpy universal function.

    The documentation of the universal function is normally the third line.
    """
    if isinstance(func, np.ufunc):

        doc = parse_docstring(func.__doc__)

        if inputs:
            func = pos_signature_modifier(func, inputs)
        else:
            raise Exception(
                f"Node {repr(node)} numpy.ufunc type function "
                "requires 'inputs' definition."
            )

        return {"_func": func, "functype": "numpy.ufunc", "doc": doc}


def model_parser(node, func, output, inputs, modifiers):
    """Check if the function is a Model class instance."""

    if isinstance(func, Model):

        doc = ""
        if func.description.splitlines():
            doc = func.description.splitlines()[0]

        if inputs:
            func = signature_modifier(func, inputs)

        return {"_func": func, "functype": "mmodel.Model", "doc": doc}


parser_engine = NodeParser(
    {
        "buitin": builtin_parser,
        "numpy.ufunc": ufunc_parser,
        "mmodel.Model": model_parser,
        "default": default_parser,
    }
)



"""Filters that used to create subgraph"""
import networkx as nx


def subnodes_by_inputs(graph, inputs: list) -> list:
    """Obtain a list of subgraph nodes based on node inputs

    If a parent node is included, so are the child nodes.

    :return: list of node names
    """

    subgraph_nodes = []

    for node, sig in nx.get_node_attributes(graph, "sig").items():
        sig_params = sig.parameters
        for param in inputs:
            if param in sig_params:
                subgraph_nodes.append(node)
                subgraph_nodes.extend(nx.descendants(graph, node))

    return subgraph_nodes


def subnodes_by_outputs(graph, outputs: list) -> list:
    """Obtain a list of subgraph nodes based on node outputs.

    For mmodel graphs, outputs from all the internal nodes are unique.
    Therefore the function only checks if function nodes overlap with
    the target return list. If a child node is included, so are the
    parent nodes.

    :return: list of node names
    """

    subgraph_nodes = []
    for node, output in nx.get_node_attributes(graph, "output").items():

        if output in outputs:

            subgraph_nodes.append(node)
            subgraph_nodes.extend(nx.ancestors(graph, node))

    return subgraph_nodes



from inspect import signature, Signature, Parameter
import networkx as nx
import textwrap

# graph properties
def modelgraph_signature(graph):
    """Obtain the signature from the model graph.

    :param DiGraph graph: networkx.Digraph() object,
        with 'signature', and 'output' defined for nodes
        and "parameters" for edges.
        The args are a dictionary of inspected signature.
    """

    parameters = {}
    for sig in nx.get_node_attributes(graph, "sig").values():
        for pname, param in sig.parameters.items():
            # remove the default values
            if param.default is param.empty and param not in parameters:
                parameters.update({pname: param})

    for output in nx.get_node_attributes(graph, "output").values():
        parameters.pop(output, None)  # if doesn't exist return None

    return Signature(sorted(parameters.values(), key=param_sorter))


def modelgraph_returns(graph):
    """Obtain the return parameter from the model graph.

    The assumption is that all return parameter names are unique.
    The function checks all returns values and all intermediate values (edge values)

    :returns: list of variable names based on note outputs
    :rtype: list
    """

    returns = []
    intermediate = []

    for node in graph.nodes():
        if graph.nodes[node]["output"]:  # skip None
            returns.append(graph.nodes[node]["output"])
    for edge in graph.edges():
        intermediate.append(graph.edges[edge]["var"])

    final_returns = list(set(returns) - set(intermediate))
    final_returns.sort()
    return final_returns


def param_sorter(parameter):
    """Sorter for argument parameter.

    The values in the tuple are compared in sequential order:
    1. Order by parameter kind
    2. Default parameter rank at the end of its kind
    3. Alphabetical order

    :param inspect.Parameter parameter: parameter object
    :rtype: (bool, parameter.name, parameter.kind)
    """

    if parameter.default is not parameter.empty:
        return parameter.kind, True, parameter.name
    else:
        return parameter.kind, False, parameter.name


def replace_signature(signature, replacement_dict):
    """Replace signature with a dictionary of (key, pair).

    The function is used to replace several input parameters with an object.
    The signature is the original signature. The dictionary key should be the
    replacement object, and the values should be a list of the target
    parameters to be replaced. The replacement allows unused parameters, they
    are skipped.
    """

    params = dict(signature.parameters)
    for func, target_list in replacement_dict.items():
        for target in target_list:
            # del params[target]
            params.pop(target, None)
        params[func] = Parameter(func, 1)

    return signature.replace(parameters=sorted(params.values(), key=param_sorter))


def graph_topological_sort(graph):
    """Determine the topological order.

    `nx.topological_generations` outputs a generator with each node list generation.
    However, it does not carry the node attributes. The method
    outputs a list of nodes for each generation.

    :return: topological order of the graph. Returns a list of nodes and its
        attribute.
    :rtype: list

    """

    topological_order = []

    for node in nx.topological_sort(graph):
        topological_order.append((node, graph.nodes[node]))

    return topological_order


def replace_subgraph(
    graph, subgraph, name, func, output=None, inputs=None, modifiers=None
):
    """Replace subgraph with a node.

    Find all parent nodes, not in the subgraph but child nodes in the
    subgraph. (All child nodes of subgraph nodes are in the subgraph).
    The edge attribute is passed down to the new edge. Here a new graph is
    created by deep copy the original graph.

    :param graph model_graph: model_graph to modify.
    :param graph subgraph: subgraph that is being replaced by a node
        subgraph is a view of the original graph.
    :param str subgraph_name: name of the subgraph.
    :param str output: output parameter name.
    """

    graph = graph.deepcopy()

    new_edges = []
    for node in subgraph.nodes():
        for parent in graph.predecessors(node):
            if parent not in subgraph:
                new_edges.append((parent, name))
        for child in graph.successors(node):
            if child not in subgraph:
                new_edges.append((name, child))

    graph.remove_nodes_from(subgraph.nodes)
    # remove unique edges
    graph.add_edges_from(set(new_edges))

    graph.set_node_object(
        name, func=func, output=output, inputs=inputs, modifiers=modifiers
    )

    return graph


def modify_node(
    graph, node, func=None, output=None, inputs=None, modifiers=None, inplace=False
):
    """Modify node.

    The result is a new graph with the node object modified.
    :param str output: change the output of the node. If the node is not
        terminal, the output should not be changed.
    :param bool inplace: if True, the original graph is modified.

    .. Note::

        If the original node has output, the node cannot be modified to None.
        If the original node has inputs, the modification cannot remove the
        original input (set to [] does not change anything).

        For the above two cases, a copy of the graph should be created and run
        ``set_node_object`` again to reset the node object.

    """
    if not inplace:
        graph = graph.deepcopy()

    func = func or graph.nodes[node]["func"]
    modifiers = modifiers or graph.nodes[node]["modifiers"]
    output = output or graph.nodes[node]["output"]
    graph.set_node_object(
        node, func=func, output=output, inputs=inputs, modifiers=modifiers
    )

    return graph


def parse_parameters(parameters):
    """Parse a list of parameters to signatures

    :param list parameters: Parameters to parse. The element can either be a string
        as the parameter name or a tuple/list as (parameter, default).
    :return: parameter order and signature.
    """

    param_order = []  # parameters in the correct order
    sig_list = []  # signature values
    sig_df_list = []  # default values
    defultargs = {}  # default arguments dictionary
    for var in parameters:
        if isinstance(var, tuple) or isinstance(var, list):
            var_name, default_value = var
            param_order.append(var_name)
            sig_df_list.append(Parameter(var_name, 1, default=default_value))
            defultargs[var_name] = default_value
        else:
            param_order.append(var)
            sig_list.append(Parameter(var, 1))

    sig = Signature(sig_list + sig_df_list)  # the default values are ordered next

    return sig, param_order, defultargs


def param_counter(graph, returns):
    """Count the number of times a parameter is used for graph execution.

    Count all function signature parameters. For extra returns,
    add one to each count value.

    :param list returns: method returns (include extra returns)
    :return: dictionary with parameter_name: count pair
    :rtype: dict
    """

    value_list = []
    for sig in nx.get_node_attributes(graph, "sig").values():
        for key, param in sig.parameters.items():

            if param.default is param.empty:
                value_list.append(key)

    # add the additional parameter to list
    value_list += returns

    count = {}
    for value in value_list:
        count[value] = count.get(value, 0) + 1

    return count


def parse_input(signature, *args, **kwargs):
    """parse argument based on signature and input.

    The default value is automatically filled.
    """

    values = signature.bind(*args, **kwargs)
    values.apply_defaults()
    return values.arguments


def content_wrap(content_list: list, width: int = 80, indent: int = 2):
    """Wrap metadata content.

    The width defaults to 80 characters. The content is a list of
    strings representing each line. The resulting wrapping has no
    initial indentation. The indent parameter is the subsequent indent
    parameter in the wrap function. The tabsize is the same as
    the indent.
    """

    wrapper = textwrap.TextWrapper(
        width=width,
        subsequent_indent=" " * indent,
        replace_whitespace=False,
        expand_tabs=True,
        tabsize=indent,
    )

    wrapped_list = []
    for item in content_list:
        if item:
            wrapped_list.extend(wrapper.wrap(item))
        else:
            wrapped_list.append("")

    return wrapped_list  # return the wrapped list


def parse_modifiers(modifiers):
    """Parse modifiers parameters to readable strings."""

    if modifiers:
        modifier_str_list = ["modifiers:"]

        for mod, kwargs in modifiers:
            str_value = [repr(v) for v in kwargs.values()]
            mod_str = f"\t- {mod.__name__}({', '.join(str_value)})"
            modifier_str_list.append(mod_str)

        return modifier_str_list
    else:
        return []


def is_node_attr_defined(graph, attr: str, attr_name: str = None):
    """Check if all graph nodes have the attribute defined.

    :param str attr: attribute string.
    :param str attr_name: the detailed name of the attribute.

    Raise an exception if the attribute is undefined.
    """
    attr_name = attr_name or attr

    if graph.name:
        graph_str = f"graph ({graph.name})"
    else:
        graph_str = "graph"

    node_list = []
    for node, node_attr in graph.nodes.data():
        if attr not in node_attr:
            node_list.append(node)

    if node_list:
        raise Exception(
            f"invalid {graph_str}: {attr_name} "
            f"{repr(attr)} is not defined for node(s) {node_list}."
        )

    return True


def is_edge_attr_defined(graph, attr: str, attr_name: str = None):
    """Check if all graph edges have the target attribute defined.

    Raise an exception if the attribute is undefined.
    """
    attr_name = attr_name or attr

    if graph.name:
        graph_str = f"graph ({graph.name})"
    else:
        graph_str = "graph"

    edge_list = []
    for u, v, edge_attr in graph.edges.data():
        if attr not in edge_attr:
            edge_list.append((u, v))
    if edge_list:
        raise Exception(
            f"invalid {graph_str}: {attr_name} {repr(attr)}"
            f" is not defined for edge(s) {edge_list}."
        )

    return True



import graphviz
from copy import deepcopy

DEFAULT_SETTINGS = {
    "graph_attr": {
        "labelloc": "t",
        "labeljust": "l",
        "splines": "ortho",
        "ordering": "out",
    },
    "node_attr": {"shape": "box"},
}


def draw_graph(G, label, style, export=None, wrap_width=30):
    """Draw a detailed graph with options.

    :param str name: name of the graph
    :param str label: title of the graph
    :param str style: there are three valid styles, plain
        short and full. Each style corresponds to node-only,
        function-only, and detailed note metadata graph
    :param str export: filename to export to
    """

    label = label.replace("\n", "\l") + "\l"
    settings = deepcopy(DEFAULT_SETTINGS)
    settings["graph_attr"].update({"label": label})

    dot_graph = graphviz.Digraph(name=G.name, **settings)

    if style == "plain":
        for node in G.nodes:
            dot_graph.node(node)

        for u, v in G.edges:
            dot_graph.edge(u, v)

    else:
        if style == "short":
            for node, ndict in G.nodes(data=True):
                if "func" in ndict:
                    nlabel = (
                        G.node_metadata(node, False, wrap_width).replace("\n", "\l")
                        + "\l"
                    )
                else:
                    nlabel = node

                dot_graph.node(node, label=nlabel)

        elif style == "full":
            for node, ndict in G.nodes(data=True):
                if "func" in ndict:
                    nlabel = (
                        G.node_metadata(node, True, wrap_width).replace("\n", "\l")
                        + "\l"
                    )
                else:
                    nlabel = node

                dot_graph.node(node, label=nlabel)

        for u, v, edict in G.edges(data=True):

            if "var" in edict:
                xlabel = edict["var"]
            else:
                xlabel = ""

            dot_graph.edge(u, v, xlabel=xlabel)

    if export:
        dot_graph.render(outfile=export)

    return dot_graph



from functools import wraps
import inspect
from mmodel.utility import parse_input, parse_parameters


def loop_modifier(func, parameter: str):
    """Modify function to iterate one given parameter.

    :param list parameter: target parameter to loop
    """

    @wraps(func)
    def loop_wrapped(**kwargs):

        loop_values = kwargs.pop(parameter)

        try:
            # make sure the parameter input is iterable
            iter(loop_values)
        except TypeError:
            raise Exception(f"{parameter} value is not iterable")

        return [func(**kwargs, **{parameter: value}) for value in loop_values]

    return loop_wrapped


def zip_loop_modifier(func, parameters: list):
    """Modify function to iterate the parameters pairwise.

    :param list parameters: list of the parameter to loop
        only one parameter is allowed. If the string of parameters is
        provided, the parameters should be delimited by ", ".
    """

    @wraps(func)
    def loop_wrapped(**kwargs):

        loop_values = [kwargs.pop(param) for param in parameters]

        result = []
        for value in zip(*loop_values):  # unzip the values

            loop_value_dict = dict(zip(parameters, value))
            rv = func(**kwargs, **loop_value_dict)
            result.append(rv)

        return result

    return loop_wrapped


def signature_modifier(func, parameters):
    """Replace node object signature.

    :param list parameters: signature parameters to replace the original
        signature. The parameters are assumed to be "POSITIONAL_OR_KEYWORD", and no
        default values are allowed. The signature will replace the original signature
        in order.

    .. Note::

        The modifier does not work with functions that have positional-only
        input parameters; or functions do not have a signature
        (built-in and numpy.ufunc). Use pos_signature_modifier instead.

    """
    sig, param_order, defaultargs = parse_parameters(parameters)

    # if there's "kwargs" in the parameter, ignore the parameter
    old_parameters = []
    for name, param in inspect.signature(func).parameters.items():
        if param.kind < 4:
            old_parameters.append(name)

    # once unzipped to return iterators, the original variable returns none
    param_pair = list(zip(old_parameters, param_order))

    @wraps(func)
    def wrapped(**kwargs):
        # assume there are no repeated signature names that are not overlapping
        # replace a, b with b, c is not allowed in the following step
        kwargs.update(defaultargs)
        for old, new in param_pair:
            kwargs[old] = kwargs.pop(new)
        return func(**kwargs)

    wrapped.__signature__ = sig
    return wrapped


def pos_signature_modifier(func, parameters):
    """Replace node object signature with position arguments.

    For functions that do not have a signature or only allow positional only
    inputs. The modifier is specifically used to change the "signature" of the builtin
    function or numpy.ufunc.

    .. note::

        The modifier is only tested against built-in function or numpy.ufunc.
    """
    sig, param_order, defaultargs = parse_parameters(parameters)

    @wraps(func)
    def wrapped(**kwargs):
        kwargs.update(defaultargs)
        # extra the variables in order
        inputs = [kwargs[key] for key in param_order]

        return func(*inputs)

    wrapped.__signature__ = sig
    return wrapped


def signature_binding_modifier(func):
    """Add parameter binding and checking for function.

    The additional wrapper is unnecessary, but to keep a consistent
    modifier syntax. The modifier can be used on wrapped functions
    that do not have a parameter binding step (ones that only allow
    keyword arguments).

    The parse_input method binds the input args and kwargs and fills
    default values automatically. The resulting function behaves the
    same as a python function.
    """

    sig = inspect.signature(func)

    @wraps(func)
    def wrapped(*args, **kwargs):

        parsed_kwargs = parse_input(sig, *args, **kwargs)

        return func(**parsed_kwargs)

    return wrapped



from collections import UserDict
from mmodel.utility import modelgraph_signature, graph_topological_sort, param_counter
from datetime import datetime
import h5py
import string
import random

ERROR_FORMAT = """\
An exception occurred for node '{node}':
--- node info ---
{node_str}
--- input info ---
{input_str}
"""


class TopologicalHandler:
    """Base class for executing graph nodes in topological order.

    "Returns" specifies the output order. If there is only one return
    the value is outputted, otherwise a tuple is outputted. This
    behavior is similar to the Python function.

    The topological handler assumes each node has exactly one output.

    :param str name: name of the handler (same as the model instance)
    :param networkx.digraph graph: graph
    :param list returns: handler returns order
        The list should have the same or more elements than the graph returns.
        See Model constructor definition.
    """

    DataClass = None

    def __init__(self, name, graph, returns: list = [], **datacls_kwargs):

        self.__name__ = name
        # __signature__ allows the inspect module to properly generate the signature
        self.__signature__ = graph.signature
        self.returns = returns
        self.order = graph_topological_sort(graph)
        self.graph = graph
        self.datacls_kwargs = datacls_kwargs

    def __call__(self, **kwargs):
        """Execute graph model by layer."""

        data = self.DataClass(kwargs, **self.datacls_kwargs)

        for node, node_attr in self.order:
            self.run_node(data, node, node_attr)

        result = self.finish(data, self.returns)

        return result

    def run_node(self, data, node, node_attr):
        """Run the individual node."""

        # Only parameters without defaults are applied

        kwargs = {
            key: data[key]
            for key, param in node_attr["sig"].parameters.items()
            if (param.default is param.empty)
        }
        try:
            # execute
            func_result = node_attr["func"](**kwargs)
            output = node_attr["output"]
            if output:  # skip the None
                data[output] = func_result

        except:  # exception occurred while running the node
            try:  # if the data class needs to be closed
                data.close()
            except:
                pass

            # format the error message
            node_str = self.graph.node_metadata(node)
            input_str = "\n".join(
                [f"{key} = {repr(value)}" for key, value in kwargs.items()]
            )
            msg = ERROR_FORMAT.format(node=node, input_str=input_str, node_str=node_str)
            raise Exception(msg)

    def finish(self, data, returns):
        """Finish execution."""

        if len(returns) == 0:
            result = None
        elif len(returns) == 1:
            result = data[returns[0]]
        else:
            result = tuple(data[rt] for rt in returns)

        try:  # if the data class needs to be closed
            data.close()
        except:
            pass

        return result


class MemData(UserDict):
    """Modified dictionary that checks the counter every time a value is accessed."""

    def __init__(self, data, counter):
        """Counter is a copy of the counter dictionary."""
        self.counter = counter.copy()
        super().__init__(data)

    def __getitem__(self, key):
        """When a key is accessed, reduce the counter.

        If the counter has reached zero, pop the value (key is deleted)
        else wise return the key.
        """

        self.counter[key] -= 1

        if self.counter[key] == 0:
            # return the value and delete the key in the dictionary
            value = super().__getitem__(key)
            del self[key]
            return value

        else:
            return super().__getitem__(key)


class H5Data:
    """Data class to interact with underlying h5 file.

    The "timestamp-uuid" is used to ensure unique entries to the H5 group.
    The randomly generated short uuid has 36^5, which is roughly 2e9
    possibilities (picoseconds range).
    """

    def __init__(self, data, fname, gname):

        self.fname = fname

        self.f = h5py.File(self.fname, "a")

        alphabet = string.ascii_lowercase + string.digits
        shortuuid = "".join(random.choices(alphabet, k=6))
        self.gname = f"{gname} {datetime.now().strftime('%y%m%d-%H%M%S')}-{shortuuid}"
        self.group = self.f.create_group(self.gname)
        self.update(data)

    def update(self, data):
        """Write key values in bulk."""
        for key, value in data.items():
            self[key] = value

    def __getitem__(self, key):
        """Read dataset/attribute by the group.

        :param str key: value name
        :param h5py.group group: open h5py group object
        """

        return self.group[key][()]

    def __setitem__(self, key, value):
        """Write h5 dataset/attribute by the group.

        If the object type cannot be recognized by HDF5, the string representation
        of the object is written as an attribute
        :param dict value_dict: the dictionary of values to write
        :param h5py.group group: open h5py group object
        """
        try:
            self.group.create_dataset(key, data=value)
        except TypeError:
            # TypeError: Object dtype dtype('O') has no native HDF5 equivalent
            self.group.attrs[key] = str(value)

    def close(self):
        """Close the data object"""
        self.f.close()


class BasicHandler(TopologicalHandler):
    """Basic handler, use the dictionary as a data class."""

    DataClass = dict


class MemHandler(TopologicalHandler):
    """Memory optimized handler, delete intermediate values when necessary.

    The process works by keeping a record of the parameter counter.
    See MemData class for more details.
    """

    DataClass = MemData

    def __init__(self, name, graph, returns: list = []):
        """Add counter to the object"""

        counter = param_counter(graph, returns)

        super().__init__(name, graph, returns, counter=counter)


class H5Handler(TopologicalHandler):
    """H5 Handler, saves all calculation values to an h5 file.

    :param str fname: h5 file name
    :param str gname: group name for the data entry
    """

    DataClass = H5Data

    def __init__(self, name, graph, returns, fname: str, gname: str = ""):
        super().__init__(name, graph, returns, fname=fname, gname=gname)



from mmodel.utility import (
    parse_input,
    parse_modifiers,
    content_wrap,
    is_node_attr_defined,
    is_edge_attr_defined,
)
from mmodel.draw import draw_graph
import networkx as nx


class Model:
    """Create model executable.

    :param str name: Model name
    :param object graph: ModelGraph instance (digraph)
    :param class handler: Handler class that handles model execution and
        the keyword arguments. The parameter format is (HandlerClass, {})
        By default, the handler takes the graph as the first parameter.
        For additional arguments, add an argument to the dictionary afterward.
    :param list modifiers: modifiers used for the whole graph model executable.
        The parameter is Optional and defaults to an empty list. For each modifier,
        the format is (modifier, {}). All modifiers should have the function as
        the first argument.
    :param str description: model description
    :param list returns: The order of returns of the model; defaults to the
        topological search.
    """

    def __init__(
        self,
        name,
        graph,
        handler,
        modifiers: list = None,
        description: str = "",
        returns: list = None,
    ):

        assert self._is_valid_graph(graph)
        self.name = self.__name__ = name

        # create a copy of the graph
        self._graph = nx.freeze(graph.deepcopy())
        self.returns = returns or self._graph.returns  # tuples
        self.modifiers = modifiers or list()
        self.handler = handler
        self.description = description

        handler_class, handler_kwargs = handler

        executor = handler_class(name, self._graph, self.returns, **handler_kwargs)

        for mdf, kwargs in self.modifiers:
            executor = mdf(executor, **kwargs)

        self.__signature__ = executor.__signature__

        # final callable
        self.executor = executor

    def __call__(self, *args, **kwargs):

        # process inputs
        inputs = parse_input(self.__signature__, *args, **kwargs)

        return self.executor(**inputs)

    def __str__(self):
        """Output callable information."""

        return self.metadata()

    def metadata(self, full=True, wrap_width=80):
        """Parse metadata string of the Model instance."""

        # use tuple if there are multiple returns
        # else use returns directly.
        return_len = len(self.returns)
        if return_len == 0:
            returns_str = "None"
        elif return_len == 1:
            returns_str = self.returns[0]
        else:
            returns_str = f"({', '.join(self.returns)})"

        metadata_list = [
            f"{self.__name__}{self.__signature__}",
            f"returns: {returns_str}",
            f"handler: {self.handler[0].__name__}"
            f"({', '.join(repr(v) for v in self.handler[1].values())})",
        ]

        metadata_list.extend(parse_modifiers(self.modifiers))

        if full:
            metadata_list.extend(["", self.description])

        wrapped_list = content_wrap(metadata_list, width=wrap_width)

        return "\n".join(wrapped_list).rstrip()

    @staticmethod
    def _is_valid_graph(G):
        """Check if the model graph is valid to build an executable.

        ``mmodel`` does not allow cycle graphs, graphs with isolated nodes,
        and all nodes have callable attributes defined.
        The method is bound to the Model class because the features
        are specific to ``Model`` class.
        """

        assert nx.is_directed(G), f"invalid graph ({G.name}): undirected graph."
        assert not nx.recursive_simple_cycles(
            G
        ), f"invalid graph ({G.name}): graph contains cycles."

        assert is_node_attr_defined(G, "func", "callable")
        # the following might occur when the node object is incorrectly constructed
        assert is_node_attr_defined(G, "output")
        assert is_node_attr_defined(G, "sig", "signature")
        assert is_edge_attr_defined(G, "var", "variable")

        return True

    @property
    def graph(self):
        """The graph attribute output a copy of the graph."""
        return self._graph.deepcopy()

    def get_node(self, node):
        """Quick access to node within the model."""

        return self._graph.nodes[node]

    def get_node_func(self, node):
        """Quick access to node base callable within the model.
        
        The function helps extract the original function within
        the node.
        """

        return self._graph.nodes[node]["_func"]

    def node_metadata(self, node, full=True, wrap_width=80):
        """View a specific node."""

        return self._graph.node_metadata(node, full=True, wrap_width=80)

    def draw(self, style="full", export=None, wrap_width=30):
        """Draw the graph of the model.

        Draws the default styled graph.

        :param str style: there are three styles, plain, short, and full.
            Plain shows nodes only, short shows part of the metadata, and
            long shows all the metadata.
        :param str export: filename to save the graph as. The file extension
            is needed.

        """

        return draw_graph(self._graph, self.metadata(), style, export, wrap_width)
